"""
GOOGLE SEARCH CONSOLE READERS MODULE
"""
import pickle
from _socket import timeout

from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from oauth2client.client import OAuth2WebServerFlow

from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler
from sdc_dp_helpers.google_search_console.config_managers import get_config
from sdc_dp_helpers.google_search_console.utils import date_range


class CustomGSCReader:
    """
    Custom Google Search Console Reader
    """

    def __init__(self, config_path=None, **kwargs):
        if config_path is not None:
            self.config = get_config(config_path, fmt="yml")
            self.credentials = kwargs["credentials"]

        if config_path is None:
            self.credentials = get_config(kwargs["credentials"], fmt="json")

        self.scopes = kwargs.get(
            "scopes",
            (
                "https://www.googleapis.com/auth/webmasters.readonly",
                "https://www.googleapis.com/auth/webmasters",
            ),
        )
        self.discovery_uri = kwargs.get(
            "discovery_uri",
            ("https://www.googleapis.com/discovery/v1/apis/customsearch/v1/rest"),
        )
        self.oauth_scope = kwargs.get(
            "oath_scope", "https://www.googleapis.com/auth/webmasters.readonly"
        )
        self.redirect_uri = kwargs.get("redirect_uri", "urn:ietf:wg:oauth:2.0:oob")

        self.data_set = []

    def generate_authentication(
        self, auth_file_location="gsc_credentials.pickle"
    ) -> None:
        """
        A user friendly method to generate a .pickle file for future authentication.
        For the first time, you would need to log in with your web browser based on
        this web authentication flow. After that, it will save your credentials in
        a pickle file. Every subsequent time you run the script, it will use the
        “pickled” credentials stored in credentials.pickle to build the
        connection to Search Console.
        """
        flow = OAuth2WebServerFlow(
            self.credentials["installed"].get("client_id"),
            self.credentials["installed"].get("client_secret"),
            self.oauth_scope,
            self.redirect_uri,
        )
        authorize_url = flow.step1_get_authorize_url()
        print(f"Go to the following link in your browser: {authorize_url}")
        code = input("Enter verification code: ").strip()
        credentials = flow.step2_exchange(code)
        with open(auth_file_location, "wb") as file_location:
            pickle.dump(credentials, file_location)

    def _get_service(self) -> build:
        """
        Makes use of the .pickle cred file to establish a webmaster connection.
        """
        with open(self.credentials, "rb") as _file:
            credentials = pickle.load(_file)

        return build(
            "searchconsole", "v1", credentials=credentials, cache_discovery=False
        )

    @request_handler(wait=1, backoff_factor=0.5)
    @retry_handler(
        exceptions=(timeout, HttpError),
        total_tries=5,
        initial_wait=60,
        backoff_factor=5,
    )
    def _query_handler(self, service, request, site_url):
        """
        Run the API request that consumes a request payload and site url.
        This separates the request with the request handler from the rest of the logic.
        """
        return service.searchanalytics().query(siteUrl=site_url, body=request).execute()

    def run_query(self):
        """
        Consumes a .yaml config file and loops through the date and url
        to return relevant data from GSC API.
        """
        service: build = self._get_service()
        start_date: str = self.config.get("start_date")
        end_date: str = self.config.get("end_date")
        dimensions: list = self.config.get("dimensions")

        print(
            f"Gathering data between given dates {start_date} and {end_date}. "
            f"Querying for Site Url: {self.config.get('site_url')}."
        )

        dimension_data_set = {}
        # split request by date to reduce 504 errors
        for dimension in dimensions:
            for date in date_range(start_date=start_date, end_date=end_date):
                print(f"Querying at date: {date} for dimension: {dimension}.")
                # run until none is returned or there is no more data in rows
                row_index = 0
                while True:
                    dim_query_set = list(dict.fromkeys(["date", dimension]))
                    response = self._query_handler(
                        service=service,
                        request={
                            "startDate": date,
                            "endDate": date,
                            "dimensions": dim_query_set,
                            "metrics": self.config.get("metrics"),
                            "type": self.config.get("type"),
                            "rowLimit": self.config.get("row_limit", 25000),
                            "startRow": row_index * self.config.get("row_limit", 25000),
                            "aggregationType": self.config.get(
                                "aggregation_type", "auto"
                            ),
                            "dimensionFilterGroups": self.config.get(
                                "dimension_filter_groups", []
                            ),
                            "dataState": self.config.get("data_state", "final"),
                        },
                        site_url=self.config.get("site_url"),
                    )

                    if response is None:
                        print("Response is None, exiting process...")
                        break
                    if "rows" not in response:
                        print("No more data in given row, moving on....")
                        break

                    # added additional data that the api does not provide
                    for row in response["rows"]:
                        dataset = {
                            "site_url": self.config.get("site_url"),
                            # keeping original search_type name for reporting
                            "search_type": self.config.get("type"),
                        }

                        # get dimension data keys and values
                        dataset.update(
                            dict(
                                zip(
                                    dim_query_set,
                                    row.get("keys", []),
                                )
                            )
                        )

                        # get metrics data
                        for metric in self.config.get("metrics", []):
                            dataset[metric] = row.get(metric)

                        if dimension not in dimension_data_set:
                            dimension_data_set[dimension] = []
                        dimension_data_set[dimension].append(dataset)
                    row_index += 1

            self.data_set.append(dimension_data_set)

        return self.data_set[0]



# pylint: disable=no-self-use,unused-import,too-few-public-methods,raise-missing-from,broad-except,too-many-instance-attributes
""""
MODULE FOR ONESIGNAL API CALL
"""
import json
import time
from datetime import datetime
from urllib.error import URLError

import pandas as pd
import requests
from dateutil import parser

from sdc_dp_helpers.api_utilities.date_managers import date_handler
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler


class CustomOneSignalReader:
    """ "One Signal Custom Reader class"""

    def __init__(self, creds_file, config_file, **kwargs):
        self._creds = load_file(creds_file, "yml")
        self._config = load_file(config_file, "yml")

        self._header = {
            "Content-Type": "application/json; charset=utf-8",
            "Authorization": f"Basic {self._creds.get('api_key')}",
            "User-Agent": "Mozilla/5.0",
        }

        self._request_session = requests.Session()
        self.offset, self.data_set = 0, []

        self._csv_wait_time = kwargs.get("csv_wait_time", 90)
        self._csv_get_attempts = kwargs.get("csv_get_attempts", 10)

    def _generate_url(self, endpoint: str):
        """
        Generate url string for authentication from given endpoints.
        :param endpoint: view_notification or csv_export
        :return: Url string.
        """
        _app_id = self._creds.get("app_id")
        if endpoint == "view_notifications":
            return f"https://onesignal.com/api/v1/notifications?app_id={_app_id}"
        if endpoint == "csv_export":
            return f"https://onesignal.com/api/v1/players/csv_export?app_id={_app_id}"

        raise ValueError(
            f"Given endpoint: {endpoint} is not supported. "
            f"Try csv_export or view_notifications."
        )

    @request_handler(wait=0.5, backoff_factor=0.5)
    def _view_notification_query_handler(self, limit: int, offset: int):
        """
        Handles the view notification request attempt.
        """
        url = self._generate_url(endpoint="view_notifications")
        return requests.get(
            url=url,
            headers=self._header,
            params={"limit": limit, "total_count": "true", "offset": offset},
            timeout=30,
        )

    @request_handler(wait=1, backoff_factor=0.5)
    def _csv_export_query_handler(self):
        """
        Handles the csv export request attempt.
        """
        url = self._generate_url(endpoint="csv_export")
        return requests.post(url=url, headers=self._header, timeout=30)

    def _get_csv_export_handler(self, response):
        """
        Attempts to gather the data from the csv url, and waits while
        it is being generated.
        :param response: Onesignal response object.
        """
        csv_url = response.get("csv_file_url", None)
        if csv_url is not None:
            attempts = 1
            while True:
                try:
                    print(f"Attempting to gather data from url: {csv_url}.")
                    data_frame: pd.DataFrame = pd.read_csv(csv_url)
                    self.data_set = data_frame.to_dict(orient="records")

                    # create a usable field to write to s3
                    # also handle timestamps that have milliseconds, it's not
                    # always there, so we need to handle both
                    for row in self.data_set:
                        try:
                            row["created_at_date"] = str(
                                datetime.strptime(
                                    row["created_at"], "%Y-%m-%d %H:%M:%S"
                                ).date()
                            ).replace("-", "")
                        except ValueError:
                            row["created_at_date"] = str(
                                datetime.strptime(
                                    row["created_at"], "%Y-%m-%d %H:%M:%S.%f"
                                ).date()
                            ).replace("-", "")

                    print("Data gathered...")
                    return self.data_set

                except URLError:
                    print(
                        f"CSV file not generated, waiting for {self._csv_wait_time} seconds. "
                        f"Attempt {attempts}/{self._csv_get_attempts}."
                    )
                    time.sleep(self._csv_wait_time)
                    if attempts >= self._csv_get_attempts:
                        raise ConnectionError(
                            f"CSV file failed to generate after {attempts} attempts. "
                            "Contact Onesignal for help."
                        )
                    attempts += 1
        else:
            raise ConnectionError(response)

    def run_query(self):
        """
        Generate a compressed CSV export of all of your current user data.
        This method can be used to generate a compressed CSV export of current user data.
        View the details of multiple notifications.
        :return: The response dataset.
        """
        _endpoint = self._config.get("endpoint")
        _limit = self._config.get("limit", 50)
        print(f"Gathering data for {_endpoint}.")

        if _endpoint == "view_notifications":
            # gather data per offset given the set of limits
            initial_response = self._view_notification_query_handler(
                limit=_limit, offset=0
            ).json()
            _total_records = int(initial_response.get("total_count"))

            for offset in range(0, _total_records, _limit):
                print(f"At offset {offset} of {_total_records}.")
                response = self._view_notification_query_handler(
                    limit=_limit, offset=offset
                ).json()
                for notification in response.get("notifications"):
                    self.data_set.append(notification)

            return self.data_set

        if _endpoint == "csv_export":
            # if connection is good, try to get the csv file from the Onesignal
            # Athena wrapper, it could take time to generate, so wait for it.
            response = self._csv_export_query_handler().json()
            self._get_csv_export_handler(response=response)
            return self.data_set

        raise ValueError(
            f"Given endpoint: {_endpoint} is not supported. "
            f"Try csv_export or view_notifications."
        )



import requests
import xmltodict
import json
from sdc_dp_helpers.api_utilities.file_managers import load_file


class XMLReader:
    """
    Custom XML Reader
    """

    def __init__(self, config_filepath, **kwargs):
        self.config_path = config_filepath

        self.config = load_file(config_filepath)

        self.url = str(self.config.get("url"))
        self.client = str(self.config.get("client"))

    def xml_to_list(self, payload: bytes) -> list:
        """
        Converts bytes to json.
        :param payload: bytes obtained from xml
        :return: json dump of list
        """
        data_dict = xmltodict.parse(payload)
        data_list = self.get_list_in_dict(data_dict)

        return data_list

    @staticmethod
    def get_list_in_dict(data: dict) -> list:
        """
        Step into dictionary repeatedly until a list is reached.
        :param data: dictionary which needs to be stepped into
        :return: first list encounter in dictionary
        """
        not_list = True

        while not_list:
            if isinstance(data, dict):
                data = data[list(data.keys())[0]]
            elif isinstance(data, list):
                not_list = False
            else:
                raise Exception("List was not found")

        return data

    def _query_handler(self) -> bytes:
        """
        Fetch XML from url
        """
        return requests.get(self.url).content

    def run_query(self) -> list:

        data_bytes = self._query_handler()

        data_llst = self.xml_to_list(data_bytes)

        return data_llst



import proto
from google.ads.googleads.client import GoogleAdsClient

from sdc_dp_helpers.api_utilities.file_managers import load_file


class CustomGAdsReader:
    """
    Custom Google Ads Reader
    """

    def __init__(self, creds_path, config_path, **kwargs):
        self.creds_path = creds_path
        self.config_path = config_path

        self.config = load_file(config_path, "yml")

        self.customer_id = str(self.config.get("customer_id"))
        self.query = str(self.config.get("query"))
        self.client = GoogleAdsClient.load_from_storage(
            path=creds_path, version=kwargs.get("version", "v8")
        )

    def run_query(self):
        """
        Query run for the GoogleAdsService endpoint.
        Use this page to build a GAQL query that selects fields from the given resource.
        This includes all of the resources, fields, segments, and metrics that are selectable when
        the resource is in the FROM clause of your GAQL query.

        Query can be built here: https://developers.google.com/google-ads/api/fields/v8/
        """
        service = self.client.get_service("GoogleAdsService")
        response = service.search_stream(customer_id=self.customer_id, query=self.query)

        data_set = []
        for batch in response:
            for row in batch.results:
                data = proto.Message.to_dict(row)
                data_set.append(data)

        return data_set



"""MODULE TO HELP WITH PROCESSING FILES"""
# pylint: disable=unused-argument
from typing import Union, Dict, Any
from pathlib import Path
import json
import yaml


def load_file(file_location: str, fmt: Union[str, None] = None) -> Dict[Any, Any]:
    """
    Gathers file data from json or yaml.
    """
    config: dict = {}
    if file_location.strip().rsplit(".", maxsplit=1)[-1] not in ["json", "yml", "yaml"]:
        raise TypeError("Wrong file type provided! Expecting only json and yaml files")

    file_location = str(file_location).strip()
    if file_location.endswith("yml"):
        with open(file_location, mode="r", encoding="utf8") as yaml_file:
            config = yaml.safe_load(yaml_file)
    if file_location.endswith("json"):
        with open(file_location, mode="r", encoding="utf8") as json_file:
            config = json.load(json_file)

    return config


def local_json_writer(
    payload: dict,
    filename: str,
    root_dir: str,
    folder_path: Union[str, None] = None,
    mode: str = "w",
) -> None:
    """WRITE DATA TO JSON OBJECT"""

    file_path = root_dir
    if folder_path:
        file_path = f"{root_dir}/{folder_path}"
        # CREATE FILE PATH IF NOT EXISTS
    Path(file_path).mkdir(parents=True, exist_ok=True)

    if filename is None or str(filename).strip() == "":
        raise ValueError("please proivde a destination filename")

    filename = f"{file_path}/{filename}.json"

    with open(f"{filename}", mode=mode, encoding="utf8") as dest_file:
        json.dump(payload, dest_file, indent=4)
        print(f"Done writting data to {filename}")



# pylint: disable=missing-function-docstring,missing-module-docstring,missing-class-docstring,unused-import,import-error,ungrouped-imports,wildcard-import,unused-argument,unused-wildcard-import
import json
import ast
import requests
from oauth2 import *
from xero.auth import OAuth2Credentials
from xero.exceptions import XeroForbidden


def token_isvalid(creds: OAuth2Credentials) -> bool:
    if not isinstance(creds, OAuth2Credentials):
        raise TypeError("creds is not an object of type OAuth2Credentials")

    return creds.get_tenants()[0]["tenantId"] is not None


def get_auth_token(client_id: str, local_token_path: str) -> OAuth2Credentials:
    """
    Consumes the client id and the previous auth processes refresh token.
    This returns an authentication token that will last 30 minutes
    to make queries the minute it is used. Or it will expire in 60 days of no use.
    The newly generated last refresh token now needs token stored for
    next use.
    PS: we receive and save the auth_token in a local dir supplied by the encapsulating project
    we never interact with s3 from inside of this class
    """
    if (not client_id) or (len(client_id) != 32):
        raise ValueError("Invalid client_id")

    with open(local_token_path, "r") as token_file:
        auth_token = ast.literal_eval(token_file.read())
    print("original token:", auth_token)
    auth_creds = OAuth2Credentials(client_id, client_secret="", token=auth_token)
    return refresh_auth_token(client_id, auth_creds, local_token_path)


def refresh_auth_token(
    client_id: str, auth_creds: OAuth2Credentials, local_token_path: str
) -> OAuth2Credentials:
    cred = {
        "grant_type": "refresh_token",
        "refresh_token": auth_creds.token["refresh_token"],  #
        "client_id": client_id,
    }
    # print(cred, auth_creds.token)
    response = requests.post(
        "https://identity.xero.com/connect/token", cred, timeout=30
    )
    auth_token = response.json()
    print("response:", auth_token)
    print("auth keys", auth_token.keys())
    err_message = auth_token.get("error")
    if err_message:
        raise Exception(err_message)

    print("Writing new Xero token to path: ", local_token_path)
    with open(local_token_path, "w") as outfile:
        outfile.write(json.dumps(auth_token))

    auth_creds = OAuth2Credentials(client_id, client_secret="", token=auth_token)
    if not token_isvalid(auth_creds):
        raise XeroForbidden(
            f"Error while trying to authenticate the refreshed token: {str(auth_creds)}"
        )
    return auth_creds


def save_auth_token(auth_token):
    """
    function to persist the latest auth token to s3
    """
    # TO DO: pull update_refresh_token from readers into here

    pass



"""
    CUSTOM FALCON READER CLASSES
"""
# pylint: disable=too-few-public-methods,import-error,unused-import,too-many-locals
import os
from typing import Generator, List
from datetime import datetime, timedelta

import requests

from sdc_dp_helpers.api_utilities.date_managers import date_range_iterator
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler


class CustomFalconReader:
    """
    Custom Falcon Reader
    """

    def __init__(self, creds_file: str, config_file=None):
        """creds_file has crecdentials config_file has api pull configs"""
        self._creds: dict = load_file(creds_file, "yml")
        self._config: dict = load_file(config_file, "yml")
        self.request_session: requests.Session = requests.Session()
        self.base_url: str = "https://api.falcon.io/"
        self.steps = 50
        self.start = 0
        self.curr_channel_ids = []

    @retry_handler(exceptions=ConnectionError, total_tries=5)
    def _get_channel_ids(self) -> List[str]:
        """
        Gather all available channel ids.
        """
        print("GET: channel ids.")
        endpoint_url = f"channels?apikey={self._creds['api_key']}"
        url = f"https://api.falcon.io/{endpoint_url}"
        response = self.request_session.get(url=url, params={"limit": "9999"})

        response_data = response.json()
        if response.status_code == 200:
            channel_ids = set()
            for item in response_data.get("items", []):
                channel_ids.add(item["id"])

            return list(channel_ids)

        raise ConnectionError(
            f"Falcon API failed to return channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=900)
    def _content_metrics_by_channel_id(
                        self, start_date: str, end_date: str, channel_ids: list
    ) -> list:
        """Gets the content metrics by channel id
        SEE: https://falconio.docs.apiary.io/ \
        #reference/channel-api/get-facebook-and-instagram-page-insights-metrics\
        # /get-content-metrics-by-channel-ids
        :param: channel_id - integer
        :param: date - string pushed as '2022-07-20'
        :returns: list of dictionaries
        """
        dataset: list = []
        endpoint_url: str = (
            f"measure/api/v1/content/metrics?apikey={self._creds['api_key']}"
        )
        offset: int = 0
        limit = self._config.get("limit", 9999)
        while True:
            print(
                f"INFO: channel id index: {self.start}, channel ids: "
                f"{len(channel_ids)}, date: {start_date}, offset: {offset}."
            )
            end_date = datetime.strptime(end_date, "%Y-%m-%d") - timedelta(days=1)
            end_date = datetime.strftime(end_date, "%Y-%m-%d")
            response = self.request_session.post(
                url=f"{self.base_url}{endpoint_url}",
                headers={"Content-Type": "application/json"},
                json={
                    "metrics": self._config.get("metrics", []),
                    "channels": channel_ids,
                    "since": start_date,
                    "until": end_date,
                    "postsSince": start_date,
                    "postsUntil": end_date,
                    "direction": self._config.get("direction", "ASC"),
                    "limit": limit,
                    "offset": offset,
                },
            )
            status_code: int = response.status_code
            reason: str = response.reason
            if status_code == 200:
                results: list = response.json()
                result_count: int = len(results)

                if result_count == 0:
                    break  # we will exit the loop because no data is got back

                if result_count < limit:
                    offset += result_count
                else:
                    offset += limit

                dataset.extend(results)
            elif status_code == 414 and "Request-URI Too Long" in reason:
                self.steps = self.steps - 20
                self.curr_channel_ids = self.curr_channel_ids[self.start : self.steps]
                print(f"len of channel_ids {len(channel_ids)}")
                raise TimeoutError(
                    "Request-URI Too Long. Reducing number of channel_ids by 20"
                )
            elif reason == "Not Found" and status_code == 404:
                print(f"No data for channel ids: {channel_ids}, skipping.\n")
                break
            elif status_code == 429:
                raise TimeoutError("Rolling Window Quota [429] reached.")
            else:
                raise ConnectionError(f"Reason: {reason}, Status: {status_code}.")

        return dataset

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=900)
    def _published_posts_by_channel_id(
                        self, start_date: str, end_date: str, channel_ids: list
    ) -> list:
        """Gets the published posts by channel id
        SEE: https://falconio.docs.apiary.io/
        #reference/content-api/get-content-metrics-by-channel-ids
        :param: channel_id - integer
        :param: date - string '2022-07-20' but converted to isoformat '2022-07-20T00:00:00'
        :returns: list of dictionaries
        """

        dataset: list = []
        endpoint_url: str = f"publish/items?apikey={self._creds['api_key']}"
        limit = self._config.get("limit", 2000)
        start_date = datetime.strptime(start_date, "%Y-%m-%d")
        end_date = datetime.strptime(end_date, "%Y-%m-%d")
        start_date = datetime.strftime(start_date, "%Y-%m-%dT%H:%M:%SZ")
        end_date = datetime.strftime(end_date, "%Y-%m-%dT%H:%M:%SZ")
        while endpoint_url:
            print(
                f"INFO: channel id index: {self.start}, "
                f"channel id: {channel_ids},  date: {start_date}, offset: {len(dataset)}."
            )
            response = self.request_session.get(
                url=f"{self.base_url}{endpoint_url}",
                headers={"Content-Type": "application/json"},
                params={
                    "channels": channel_ids,
                    "since": start_date,
                    "until": end_date,
                    "networks": self._config["networks"],
                    "statuses": self._config.get("statuses", "published"),
                    "limit": limit,
                },
            )
            status_code: int = response.status_code
            reason: str = response.reason
            if response.status_code == 200:
                results: dict = response.json()
                items_data: list = results.get("items", [])
                dataset.extend(items_data)

                endpoint_url = results.get("next", {"href": None}).get("href")

                if len(items_data) == 0 or endpoint_url is None:
                    break

            elif reason == "Not Found" and status_code == 404:
                print(f"No data for channel id: {channel_ids}, skipping.\n")
                break
            elif status_code == 429:
                raise TimeoutError("Rolling Window Quota [429] reached.")
            else:
                raise ConnectionError(f"Reason: {reason}, Status: {status_code}.")

        return dataset

    def _query_handler(self, start_date: str, end_date: str, channel_ids: list) -> list:
        """Interface to decide the api call to make"""
        endpoint_name = self._config.get("endpoint_name", None)
        self.start = 0
        dataset = []
        if not endpoint_name:
            raise KeyError(
                "Please specify endpoint in the configs i.e content_metrics or published_posts"
            )
        if endpoint_name == "content_metrics":
            channel_len = len(channel_ids)
            for _ in range(0, channel_len, self.steps):
                self.curr_channel_ids = channel_ids[
                    self.start : self.start + self.steps
                ]
                results = self._content_metrics_by_channel_id(
                    start_date, end_date, self.curr_channel_ids
                )
                dataset.extend(results)
                self.start += self.steps
        elif endpoint_name == "published_posts":
            for channel_id in channel_ids:
                self.curr_channel_ids = [channel_id]
                results = self._published_posts_by_channel_id(
                    start_date, end_date, [channel_id]
                )
                dataset.extend(results)
                self.start += 1
        else:
            raise ValueError(f"Invalid endpoint_name {endpoint_name}")

        return dataset

    def run_query(self) -> Generator[dict, None, None]:
        """
        Get metrics by channel Id context returns a request session with the ability
        to page with offsets.
        Content (or Post level) contains all metrics about your specific piece of
        content (posts). Here you will find impressions, reach, likes,
        shares and other metrics that show how well your specific post has performed.
        https://falconio.docs.apiary.io/reference/content-api/get-copied-content.
        """
        channel_ids: list = self._get_channel_ids()[:100]
        # channel_ids = ["114215121990330", "58283981268"]
        print(f"Attempting to gather metrics from {len(channel_ids)} channel ids.")
        date_iterator = date_range_iterator(
            start_date=self._config["since"],
            end_date=self._config["until"],
            interval="1_day",
            end_inclusive=False,
            time_format="%Y-%m-%d",
        )
        for start_date, end_date in date_iterator:
            payload: list = []
            payload = self._query_handler(
                start_date=start_date, end_date=end_date, channel_ids=channel_ids
            )
            if len(payload) > 0:
                yield {
                    "networks": self._config["networks"],
                    "date": start_date,
                    "data": payload,
                }
            else:
                print(
                    f"No data for endpoint {self._config['endpoint_name']} for date : {start_date}"
                )


class CustomFalconReaderV2:
    """
    Custom Falcon Reader for Content Insights
    """

    def __init__(self, creds_file: str, config_file=None):
        """creds_file has crecdentials config_file has api pull configs"""
        self._creds: dict = load_file(creds_file, "yml")
        self._config: dict = load_file(config_file, "yml")
        self.request_session: requests.Session = requests.Session()
        self.base_url: str = "https://api.falcon.io/"
        self.success = []
        self.start = 0
        self.curr_channels = []

    def is_success(self):
        """Append True to Success List for successfull API pull"""
        self.success.append(True)

    def not_success(self):
        """Append False for failed API pull"""
        self.success.append(False)

    def _get_channel_ids(self) -> dict:
        """
        Gather all available channel ids.
        """
        print("GET: channel ids.")
        endpoint_url = f"channels?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        response = self.request_session.get(url=url, params={"limit": "9999"})

        response_data = response.json()
        if response.status_code == 200:
            channel_ids = {}
            for item in response_data.get("items", []):
                channel_ids[item["uuid"]] = item["name"]
            # print(len(channel_ids))
            return channel_ids

        raise ConnectionError(
            f"Falcon API failed to return channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )

    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def _get_content_ids_by_channel_id(
                        self, start_date: str, end_date: str, network: str, channel_ids: List[str]
    ) -> dict:
        """Retrieves all social media published
        SEE: https://falconio.docs.apiary.io/
        #reference/content-api/get-channel-content
        :param: date - string '2022-07-20' but converted to isoformat '2022-07-20T00:00:00Z'
        :returns: list of dictionaries
        """
        print("GET: content ids by channel id.")
        date_filters = f"since={start_date}&until={end_date}"
        endpoint_url = (
            f"publish/items?apikey={self._creds['api_key']}"
            f"&statuses=published"
            f"&networks={network}"
            f"&{date_filters}"
        )
        content_ids_by_channel_id = {}
        while endpoint_url is not None:
            url = f"{self.base_url}{endpoint_url}"
            response = self.request_session.get(url=url, params={"limit": "9999"})
            response_data = response.json()
            if response.status_code != 200:
                raise ConnectionError(
                    f"Falcon API failed to return content ids. "
                    f"Status code: {response.status_code}, Reason: {response.reason}."
                )
            for item in response_data.get("items"):
                content_id = item.get("id")
                channel_id = item.get("channels")
                if channel_id is not None and channel_id[0] in channel_ids:
                    content_ids_by_channel_id.setdefault(channel_id[0], []).append(
                        content_id
                    )
            endpoint_url = response_data.get("next", {"href": None}).get("href")
        print(content_ids_by_channel_id)
        return content_ids_by_channel_id

    def _get_insights_request_id(
                        self, start_date: str, end_date: str, content_ids_by_channel_id: dict
    ) -> dict:
        """Get Insights request id
        channels per request : max 15
        time between since and until : max 3 months
        metricids per request : max 20
        contentids per request: max 300
        SEE: https://falconio.docs.apiary.io/
        #reference/measure-api-v2/request-content-insights
        :param: date - string '2022-07-20' but converted to isoformat '2022-07-20T00:00:00'
        :returns: dictionary
        """
        print("GET: insights request id.")
        endpoint_url = f"measure/v2/insights/content?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        body = {
            "since": start_date,
            "until": end_date,
            "metricIds": self._config["metrics"],
            "channels": [],
        }

        for key, value in content_ids_by_channel_id.items():
            body["channels"].append({"id": key, "contentIds": value})
        response = self.request_session.post(
            url=url, headers={"Content-Type": "application/json"}, json=body
        )
        if response.status_code == 200:
            response_data = response.json()
            insights_request_id = response_data["insightsRequestId"]
            return insights_request_id

        raise ConnectionError(
            f"Falcon API failed to return content ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )

    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def _get_content_insights(self, insights_request_id: str):
        """Get Insights
        It is recommended to wait 3 seconds and then check for the availability
        of the data. If the request is still in progress then check every
        3 seconds.
        SEE: https://falconio.docs.apiary.io/
        #reference/measure-api-v2/request-content-insights
        :param: date - string '2022-07-20' but converted to isoformat '2022-07-20T00:00:00'
        :returns: dictionary
        """
        print("GET: content insights.")
        endpoint_url = (
            f"measure/v2/insights/{insights_request_id}?apikey={self._creds['api_key']}"
        )
        url = f"{self.base_url}{endpoint_url}"
        response = self.request_session.get(url=url)

        response_data = response.json()
        if response.status_code == 200:
            if response_data["status"] == "READY":
                return response_data
            if response_data["status"] == "EXPIRED":
                return None
        raise ConnectionError(
            f"Falcon API failed to return content ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}."
        )

    @staticmethod
    def _normalize_data(dataset: dict, network: str, channel_ids: dict):
        normalized_data = []
        for metric, data_items in dataset.items():
            # print(metric)
            for data_item in data_items:
                data_item[f"{metric}"] = data_item.pop("value")
                data_item["network"] = network
                data_item["brand"] = channel_ids.get(data_item["channelId"], None)
                normalized_data.append(data_item)
        return normalized_data

    def _query_handler(
                        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> dict:
        endpoint_name = self._config.get("endpoint_name", None)
        if not endpoint_name:
            raise KeyError(
                "Please specify endpoint in the configs i.e content_insights"
            )
        content_ids_by_channel_id = self._get_content_ids_by_channel_id(
            start_date=start_date,
            end_date=end_date,
            network=network,
            channel_ids=channel_ids,
        )
        channels = list(content_ids_by_channel_id.keys())
        channel_id_length = len(channels)
        channel_steps = self._config.get("channel_steps", 10)
        dataset = {}
        data = []
        for _ in range(0, channel_id_length, channel_steps):
            self.curr_channels = channels[self.start : self.start + channel_steps]
            request_content_ids_by_channel_id = {
                key: value
                for key, value in content_ids_by_channel_id.items()
                if key in self.curr_channels
            }

            if len(request_content_ids_by_channel_id) > 0:
                insights_request_id = self._get_insights_request_id(
                    start_date=start_date,
                    end_date=end_date,
                    content_ids_by_channel_id=request_content_ids_by_channel_id,
                )
                results = self._get_content_insights(insights_request_id)
                if results is None:
                    insights_request_id = self._get_insights_request_id(
                        start_date=start_date,
                        end_date=end_date,
                        content_ids_by_channel_id=request_content_ids_by_channel_id,
                    )
                    results = self._get_content_insights(insights_request_id)
                data = results["data"]["insights"]
            dataset.update(data)
            self.start = self.start + channel_steps
        return self._normalize_data(dataset, network, channel_ids)

    def run_query(self):
        """Handles the query results"""
        channel_ids = self._get_channel_ids()
        print(f"Attempting to gather metrics from {len(channel_ids)} channel ids.")
        network = self._config.get("networks")
        date_iterator = date_range_iterator(
            start_date=self._config["since"],
            end_date=self._config["until"],
            interval="1_day",
            end_inclusive=False,
            time_format="%Y-%m-%dT%H:%M:%SZ",
        )
        for start_date, end_date in date_iterator:
            payload: list = []
            payload = self._query_handler(
                start_date=start_date,
                end_date=end_date,
                network=network,
                channel_ids=channel_ids,
            )
            date = start_date[:-10]
            if len(payload) > 0:
                yield {
                    "networks": network,
                    "date": date.replace("-", ""),
                    "data": payload,
                }
                self.is_success()
            else:
                self.not_success()
                print(
                    f"No data for endpoint {self._config['endpoint_name']} for date : {date}"
                )
